use crate::prelude::*;
use notify::{Config, Event, RecommendedWatcher, RecursiveMode, Watcher};
use std::path::Path;
use tokio::{sync::mpsc::channel, task::JoinHandle};
use tracing::{error, info};

pub type WatcherType = RecommendedWatcher;
pub type FileWatcher = (WatcherType, JoinHandle<()>);

pub mod prelude {
    #![allow(unused_imports)]
    pub use super::{file_monitor, file_watcher, FileWatcher};
}

/// Notifies your callback for each individual event
pub fn file_watcher<P, F>(dir: P, cb: F) -> Result<FileWatcher>
where
    P: AsRef<Path>,
    F: Fn(Event) -> () + std::marker::Send + 'static,
{
    // TODO: debounce?
    let (tx, mut rx) = channel(1);
    let mut watcher = RecommendedWatcher::new(
        move |res| match res {
            Ok(e) => futures::executor::block_on(async {
                tx.send(e).await.unwrap();
            }),
            Err(e) => error!("Error from file_watcher: {e}"),
        },
        Config::default(),
    )?;

    info!("Watching directory '{}'", dir.as_ref().display());

    let handle = tokio::spawn(async move {
        while let Some(ev) = rx.recv().await {
            cb(ev)
        }
    });

    watcher.watch(dir.as_ref(), RecursiveMode::Recursive)?;

    Ok((watcher, handle))
}

/// Only know when something changes
pub fn file_monitor<P, F>(dir: P, cb: F) -> Result<FileWatcher>
where
    P: AsRef<Path>,
    F: Fn() -> () + std::marker::Send + 'static,
{
    let (tx, mut rx) = tokio::sync::watch::channel(());
    let mut watcher = RecommendedWatcher::new(
        move |res| match res {
            Ok(_e) => futures::executor::block_on(async {
                tx.send_replace(());
            }),
            Err(e) => error!("Error from file_watcher: {e}"),
        },
        Config::default(),
    )?;

    info!("Watching directory '{}'", dir.as_ref().display());

    let handle = tokio::spawn(async move {
        while let Ok(_) = rx.changed().await {
            cb();
            // "good enough" debouncing
            rx.mark_unchanged();
        }
    });

    watcher.watch(dir.as_ref(), RecursiveMode::Recursive)?;

    Ok((watcher, handle))
}
